%{
#include <zasm/parser.h>
#include <zasm/stream.h>
#include <zasm/feeder.h>
#include <zasm/expr.h>
#include <zasm/functions.h>
  static struct expr_alloc_t *alloc = NULL;
  static struct expr_t *expr;
  static struct token_t *last_token;
  static int depth = 0;
  static struct expr_t *expr_new_number(int num)
    {expr = expr_new(&alloc); expr->type = EXPR_NUMBER; expr->value.value = num; return expr;}
  static struct expr_t *expr_new_name(char *name, char *part)
    {expr = expr_new(&alloc); expr->type = EXPR_NAME; expr->value.name.name = name; expr->value.name.part = part; return expr;}
  static struct expr_t *expr_new_operator(int (*calc)(int a, int b, int c),
					  struct expr_t *a, struct expr_t *b, struct expr_t *c)
    {expr = expr_new(&alloc); expr->type = EXPR_OPERATOR; 
    expr->value.function.calc = calc; expr->value.function.arg1 = a;
    expr->value.function.arg2 = b; expr->value.function.arg3 = c;
    return expr;}
  static struct feeder_t *feeder;
  static int yylex();
  static void yyerror(char *msg);
#define YYSTYPE struct expr_t *
%}

%token NUM
%token NAME
%left '?' ':'
%left L_OR L_AND L_XOR L_NOT
%left B_OR B_AND B_XOR B_NOT
%left S_LEFT S_RIGHT
%left EQU NEQ LT GT LE GE CMP
%left '+' '-'
%left '*' '/' '%'
%left NEG
%right EXP
%token '(' ')'

%%
final_expr: expr {expr = $1;}
 
expr : NUM  {$$ = $1;}
     | NAME {$$ = $1;} 
     | expr '?' expr ':' expr {$$ = expr_new_operator(function_ifthenelse, $1, $3, $5);}
     | expr L_OR  expr {$$ = expr_new_operator(function_log_or , $1, $3, NULL);}
     | expr L_AND expr {$$ = expr_new_operator(function_log_and, $1, $3, NULL);}
     | expr L_XOR expr {$$ = expr_new_operator(function_log_xor, $1, $3, NULL);}
     | L_NOT expr      {$$ = expr_new_operator(function_log_not, $2, NULL, NULL);}
     | expr B_OR  expr {$$ = expr_new_operator(function_bit_or , $1, $3, NULL);}
     | expr B_AND expr {$$ = expr_new_operator(function_bit_and, $1, $3, NULL);}
     | expr B_XOR expr {$$ = expr_new_operator(function_bit_xor, $1, $3, NULL);}
     | B_NOT expr      {$$ = expr_new_operator(function_bit_not, $2, NULL, NULL);}
     | expr S_LEFT  expr {$$ = expr_new_operator(function_shift_left , $1, $3, NULL);}
     | expr S_RIGHT expr {$$ = expr_new_operator(function_shift_right, $1, $3, NULL);}
     | expr EQU  expr {$$ = expr_new_operator(function_equ, $1, $3, NULL);}
     | expr NEQ expr  {$$ = expr_new_operator(function_neq, $1, $3, NULL);}
     | expr LT expr   {$$ = expr_new_operator(function_lt , $1, $3, NULL);}
     | expr GT  expr  {$$ = expr_new_operator(function_gt , $1, $3, NULL);}
     | expr LE expr   {$$ = expr_new_operator(function_le , $1, $3, NULL);}
     | expr GE expr   {$$ = expr_new_operator(function_ge , $1, $3, NULL);}
     | expr CMP expr  {$$ = expr_new_operator(function_cmp, $1, $3, NULL);}
     | expr '+' expr {$$ = expr_new_operator(function_add, $1, $3, NULL);}
     | expr '-' expr {$$ = expr_new_operator(function_sub, $1, $3, NULL);}
     | expr '*' expr {$$ = expr_new_operator(function_mul, $1, $3, NULL);}
     | expr '/' expr {$$ = expr_new_operator(function_div, $1, $3, NULL);}
     | expr '%' expr {$$ = expr_new_operator(function_mod, $1, $3, NULL);}
     | '-' expr %prec NEG {$$ = expr_new_operator(function_neg, $2, NULL, NULL);}
     | '+' expr %prec NEG {$$ = $2;}
     | expr EXP expr {$$ = expr_new_operator(function_pow, $1, $3, NULL);}
     | '(' expr ')' {$$ = $2;}
;

%%

static int yylex()
{
  while (last_token->type == TOK_SPACE)
    *last_token = feeder_get(feeder);
  switch (last_token->type)
    {
    case TOK_NAME:
      yylval = expr_new_name(last_token->value.t_char_p, NULL);
      *last_token = feeder_get(feeder);
      while (last_token->type == TOK_SPACE)
	*last_token = feeder_get(feeder);
      if (last_token->type == TOK_DOT)
	{
	  *last_token = feeder_get(feeder);
	  while (last_token->type == TOK_SPACE)
	    *last_token = feeder_get(feeder);
	  if (last_token->type != TOK_NAME)
	    T_ERROR(*last_token, "syntax error\n");
	  yylval->value.name.part = last_token->value.t_char_p;
	  *last_token = feeder_get(feeder);
	}
      return NAME;
    case TOK_INTEGER:
      yylval = expr_new_number(last_token->value.t_int);
      *last_token = feeder_get(feeder);
      return NUM;
    case TOK_OP_IFTHEN: *last_token = feeder_get(feeder); return '?';
    case TOK_OP_ELSE: *last_token = feeder_get(feeder); return ':';
    case TOK_OP_LOG_OR : *last_token = feeder_get(feeder); return L_OR;
    case TOK_OP_LOG_AND: *last_token = feeder_get(feeder); return L_AND;
    case TOK_OP_LOG_XOR: *last_token = feeder_get(feeder); return L_XOR;
    case TOK_OP_LOG_NOT: *last_token = feeder_get(feeder); return L_NOT;
    case TOK_OP_BIN_OR : *last_token = feeder_get(feeder); return B_OR;
    case TOK_OP_BIN_AND: *last_token = feeder_get(feeder); return B_AND;
    case TOK_OP_BIN_XOR: *last_token = feeder_get(feeder); return B_XOR;
    case TOK_OP_BIN_NOT: *last_token = feeder_get(feeder); return B_NOT;
    case TOK_OP_SHIFT_LEFT : *last_token = feeder_get(feeder); return S_LEFT;
    case TOK_OP_SHIFT_RIGHT: *last_token = feeder_get(feeder); return S_RIGHT;
    case TOK_OP_EQU: *last_token = feeder_get(feeder); return EQU;
    case TOK_OP_NEQ: *last_token = feeder_get(feeder); return NEQ;
    case TOK_OP_LT: *last_token = feeder_get(feeder); return LT;
    case TOK_OP_GT: *last_token = feeder_get(feeder); return GT;
    case TOK_OP_LE: *last_token = feeder_get(feeder); return LE;
    case TOK_OP_GE: *last_token = feeder_get(feeder); return GE;
    case TOK_OP_CMP: *last_token = feeder_get(feeder); return CMP;
    case TOK_OP_ADD: *last_token = feeder_get(feeder); return '+';
    case TOK_OP_SUB: *last_token = feeder_get(feeder); return '-';
    case TOK_OP_MUL: *last_token = feeder_get(feeder); return '*';
    case TOK_OP_DIV: *last_token = feeder_get(feeder); return '/';
    case TOK_OP_MOD: *last_token = feeder_get(feeder); return '%';
    case TOK_OP_POW: *last_token = feeder_get(feeder); return EXP;
    case TOK_PAREN_OPEN: depth++; *last_token = feeder_get(feeder); return '(';
    case TOK_PAREN_CLOSE: depth--; if (depth < 0) return 0; *last_token = feeder_get(feeder); return ')';
    default:
      return 0;
    }
}

static void yyerror(char *msg)
{
  T_ERROR(*last_token, "%s\n", msg);
}

struct expr_t *parse_expr(struct feeder_t *feed, struct token_t *token, struct expr_alloc_t **all)
{
  depth = 0;
  feeder = feed;
  last_token = token;
  alloc = *all;
  if (yyparse())
    ERROR("parse error\n");
  *all = alloc;
  alloc = NULL;
  return expr;
}
